package tsp.csc4509.tpnote;

import static tsp.csc4509.common.Log.COMM;
import static tsp.csc4509.common.Log.GEN;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.StandardSocketOptions;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import tsp.csc4509.tcpnio.FullDuplexMsgWorker;
import tsp.csc4509.tcpnio.ReadMessageStatus;
import tsp.csc4509.common.NetTools;
import java.nio.channels.SelectionKey;
import tsp.csc4509.debug.DebugTools;
/**
 * Classe représantant un noeud du graphe.
 * 
 * @author Éric Lallet
 *
 */
public class Node {
	/**
	 * Id du noeud qui sert de paramètre au demande de la ressource critique.
	 */
	private NodeId id;

	/**
	 * id du node père.
	 */
	private NodeId father;

	/**
	 * connexion au noeud père dans l'arbre des noeuds. Évolue au cours du temps.
	 */
	private FullDuplexMsgWorker fatherConnexion;
	
	
	/**
	 * id du noeud suivant dans la liste des noeuds an attente de la ressource.
	 */
	private NodeId next;
	
	/**
	 * connexion au noeud suivant dans la liste des noeuds en attente de la ressource critique.
	 */
	private FullDuplexMsgWorker nextConnexion;

	/**
	 * vrai si ce noeud possède le jeton.
	 * si en plus le boolean requesting est vrai, ce noeud possède la ressouce et est entrée dans la section critique;
	 * mais si le boolean requesting est faux, ce noeud doit transmettre le jeton au premier noeud qui demandera
	 * la ressource et doit alors repasser ce boolean à faux.
	 */
	private boolean havingToken;

	/**
	 * vrai si le noeud veut la ressource.
	 * si en plus le boolean havingToken est vrai, ce noeud possède la ressouce et est entrée dans la section critique.
	 */
	private boolean requesting;
	
	
	/**
	 * Canal de connexion pour les noeuds fils de l'arbre et le noeud précédant de la liste d'attente.
	 */
	private ServerSocketChannel	listenChannel;


	/**
	 * Selector pour gérer les connexions des noeuds fils, et les messages pouvant arriver de tous
	 * les noeuds voisins.
	 */
	private Selector selector;	


	/**
	 * map de toutes les connexions où il faudra faire des lectures asynchrones.
	 *    - les noeuds fils dans l'arbre (lecture des ResourceRequest)
	 *    - le noeud précedant dans la liste des noeuds en attente de la ressource (lecture d'un Token)
	 *    - la console (lecture d'un TaskResquest)
	 *    - le serveur ressource (lecrure d'un TaskResponse)
	 */
	private Map<SelectionKey, FullDuplexMsgWorker> inConnexions;


	/** 
	 * La tâche que veut réaliser ce noeud.
	 */
	private TaskRequest taskRequest;
	
	/**
	 * adresse TCP/IP du serveur ressource.
	 */
	private InetSocketAddress serverAddr;
	
	
	/**
	 * connexion au programme qui va piloter ce noeud.
	 */
	private FullDuplexMsgWorker console;
	
	

	
	/**
	 * Constructeur destiné au premier noeur racine de l'arbre.
	 *  @param num
	 *          numéro du noeud.
	 * @param myPort
	 *            port occupé par le noeud.
	 * @param serverHost
	 *            nom de la machine où tourne le serveur ressource.
	 * @param serverPort
	 *            port du serveur ressource.
	 * @throws IOException
	 *            toutes les IOException qui peuvent se produire lors de la création de la partie serveur du
	 *            noeud ou lors de la recherche des adresses IP de la machine où tourne ce noeud.
	 */
	public Node(final int num, final int myPort, final String serverHost, final int serverPort) throws IOException {
		GEN.info("Création du noeud (" + num + ")");
		InetAddress addr = NetTools.giveANoneLoopbackAddr();
		if (addr == null) {
			throw new IOException("Aucune adresse IP trouvée sur cette machine.");
		}
		
		this.serverAddr = new InetSocketAddress(InetAddress.getByName(serverHost), serverPort);
		// création du Selector pour gérer tous les évènements réseaux asynchrones (connexions, lectures)
		this.selector = Selector.open();
		this.inConnexions = new HashMap<SelectionKey, FullDuplexMsgWorker>();

		// prépare le canal pour la connexion des noeuds fils.
		openListenChannel(myPort);

		// réglage des paramètres du protocole pour le noeud racine de l'arbre de départ.
		// le constructeurs des noeuds fils doit rectifier la valeur de father, et havingToken
		this.id = new NodeId(addr, myPort, num);
		GEN.info("[" + this.id.getNum() + "] Le noeud reçoit l'id " + id);
		this.father = null;
		this.fatherConnexion = null;
		this.next = null;
		this.nextConnexion  = null;
		this.havingToken = true;
		this.requesting = false;

	}

	/**
	 * Constructeur destiné à tous les noeuds fils de l'atbre.
	 * @param myNum
	 *          numéro du noeud.
	 * @param myPort
	 *            port occupé par le noeud.
	 * @param serverHost
	 *            nom de la machine où tourne le serveur ressource.
	 * @param serverPort
	 *            port du serveur ressource.
	 * @param fatherNum
	 *            numero du noeud père.
	 * @param fatherHost
	 *            nom de la machine où tourne le noeud père de ce noeud.
	 * @param fatherPort
	 *            port TCP occupé par le noeud père de ce noeud.
	 * @throws IOException
	 *            toutes les IOException qui peuvent se produire lors de la création de la partie serveur du
	 *            noeud ou lors de la recherche des adresses IP de la machine où tourne ce noeud.
	 *            
	 */
	public Node(final int myNum, final int myPort, final String serverHost, final int serverPort, final int fatherNum, final String fatherHost, final int fatherPort) throws IOException {
		// on utilise le prmier constructeur pour la partie commune du code. 
		// il faut ensuite changer deux informations; father et havingToken qui diffèrent pour les noeuds fils.
		this(myNum, myPort, serverHost, serverPort);
		// au début, c'est le noeud racine qui possède le jeton et pas les noeuds fils.
		this.havingToken = false;
		// connexion au noeud pére
		this.father = new NodeId(new InetSocketAddress(fatherHost, fatherPort), fatherNum);
		this.fatherConnexion = connectTo(fatherHost, fatherPort);
		GEN.info("[" + this.id.getNum() + "]Le noeud se connecte au père " + fatherNum);
	}
	
	
	


	/**
	 * Crée un FullDuxplexMsgWorker connecté à l'adresse TCP/IP fournie par addr.
	 * @param addr
	 *               InetAddress (adresse TCP/IP) de connexion.
	 * @return
	 *               référence sur le FullDuplexMsgWorker connecté au serveur occupant l'adresse addr.
	 * @throws IOException
	 *               toutes les IOException qui peuvent se produire lors de la connexion.
	 */
	public static FullDuplexMsgWorker connectTo(final InetSocketAddress addr) throws IOException {
		SocketChannel rwChan = SocketChannel.open(addr);
		return new FullDuplexMsgWorker(rwChan);
	}


	/**
	 * Crée un FullDuplexMsgWorker connecté à l'adresse host:port.
	 * @param host
	 *               nom de la machine où tourne le serveur à connecter.
	 * @param port
	 *               port occupé par le serveur à connecter.
	 * @return
	 *               référence sur le FullDuplexMsgWorker connecté au serveur tournant sur host:port.
	 * @throws IOException
	 *               toutes les IOException qui peuvent se produire lors de la connexion au noeud père.
	 */
	public static FullDuplexMsgWorker connectTo(final String host, final int port) throws IOException {
		InetSocketAddress addr =  new InetSocketAddress(InetAddress.getByName(host), port);
		return connectTo(addr);

	}
	
	/**
	 * Ferme un FullDuplexMsgWorker.
	 * @param worker
	 *          la référence du FullDuplexMsgWorker. La référence peut être null, ou le worker déjà fermé.
	 * @throws IOException
	 *          toutes les IOException pouvant se produire durant la fermeture d'un canal.
	 */
	public static void closeFullDuplexMsgWorker(FullDuplexMsgWorker worker) throws IOException {
		if (worker != null && worker.getChannel().isOpen()) {
			worker.close();
		}
	}
	

	// Question 3 (2 pts)	
	/**
	 * Connexion au serveur ressource, et envoi de la requête.
	 * Le worker connecté au server est passé en aysnchrone et ajouté dans 
	 * la map des inConnexion, pour gérer la futur réponse du serveur.
	 * @throws IOException
	 * 				toutes les IOException qui peuvent se produire lors des connexions et les envois
	 */
	public void sendTaskToServer() throws IOException {
		SocketChannel rwChan = SocketChannel.open(serverAddr); //0.5
		rwChan.configureBlocking(false); 
		SelectionKey newKey = rwChan.register(selector, SelectionKey.OP_READ); // 0.5 + conf
		FullDuplexMsgWorker	worker = new FullDuplexMsgWorker(rwChan);
		inConnexions.put(newKey, worker); // 0.5
		worker.sendMsg(MessType.TASKREQUESTTYPE.ordinal(), taskRequest); // 0.5
	}
	
	// question 4 (3 pts)
	/**
	 * Mise en oeuvre de la partie de l'algorithme de Naimi-Tréhel qui gère la réception du ResourceRequest par un Node.
	 * @param resourceRequest
	 *                le ResourceRequest émis par un Node de l'arbre. La méthode resourceRequest.getNodeId() permet de
	 *                connaître l'id de ce Node, et la méthode resourceRequest.getNodeId().getAddr() permet de connaître
	 *                l'addresse TCP/IP de ce node.
	 * @throws IOException
	 *                toutes les IOException pouvant se produire lors des échanges et des connexions.
	 */
	public void receivingResourceRequest(final ResourceRequest resourceRequest) throws IOException {
		GEN.info("[" + this.id.getNum() + "] le noeud  reçoit la demande de ressources de " + resourceRequest.getNodeId().getNum());
		if (fatherConnexion == null) { // tous les tests 0.5
			if (requesting) {
				GEN.info("[" + this.id.getNum() + "] le noeud  dernier de la file d'attente, ajoute  " + resourceRequest.getNodeId().getNum() + " dans la file ");
				closeFullDuplexMsgWorker(nextConnexion);
				next = resourceRequest.getNodeId();
				nextConnexion = connectTo(resourceRequest.getNodeId().getAddr()); 
				GEN.info("[" + this.id.getNum() + "] le noeud  racine devint fils de  " + resourceRequest.getNodeId().getNum());
				father = resourceRequest.getNodeId();
				fatherConnexion = connectTo(resourceRequest.getNodeId().getAddr()); // rechainages 0.5
			} else {
				GEN.info("[" + this.id.getNum() + "] le noeud racine sans ressource donne racine et jeton à  " + resourceRequest.getNodeId().getNum());
				havingToken = false;
				GEN.info("[" + this.id.getNum() + "] le noeud  racine devint fils de  " + resourceRequest.getNodeId().getNum());
				father = resourceRequest.getNodeId();
				fatherConnexion = connectTo(resourceRequest.getNodeId().getAddr()); // rechainage 0.5
				fatherConnexion.sendMsg(MessType.TOKENTYPE.ordinal(), new Token());  // 0.5
			}
		} else {
			GEN.info("[" + this.id.getNum() + "] le noeud pas racine transmet la demande de ressource à son pere (" + father.getNum() 
			 + ") et change de père pour " + resourceRequest.getNodeId().getNum());
			fatherConnexion.sendMsg(MessType.RESOURCEREQUESTTYPE.ordinal(), resourceRequest); // 0.5
			closeFullDuplexMsgWorker(fatherConnexion);
			father = resourceRequest.getNodeId();
			fatherConnexion = connectTo(resourceRequest.getNodeId().getAddr()); // rechainage 0.5
		}
	}
	
	// question 5 (2 pts)
	/**
	 * Mise en oeuvre de la partie de l'algorithme de Naimi-Tréhel qui gère le passage d'un Node dans
	 * la liste des Node voulant la ressource.
	 * @param taskRequest
	 *             la tâche émise par la Console qui a provoqué la demande de ressource.
	 * @throws IOException
	 *             toutes les IOException pouvant se produire lors des échanges et des connexions.
	 */
	public void requestingResource(final TaskRequest taskRequest) throws IOException {
		GEN.info("[" + this.id.getNum() + "] le noeud  demande la ressource ");
		requesting = true;
		this.taskRequest = taskRequest; // 0.5
		if (fatherConnexion != null) { // test 0.5
			GEN.info("[" + this.id.getNum() + "] le noeud  envoie sa demande de ressource à son père " + father.getNum() + " et se prépare à devenir racine");
			ResourceRequest resourceRequest = new ResourceRequest(id);
			fatherConnexion.sendMsg(MessType.RESOURCEREQUESTTYPE.ordinal(), resourceRequest); // 0.5
			closeFullDuplexMsgWorker(fatherConnexion);
			father = null;
			fatherConnexion = null; // chainage 0.5
		} else {
				GEN.info("[" + this.id.getNum() + "] le noeud  est racine et en plus a déjà la jeton: il obtient la ressource");
				sendTaskToServer(); 
		}
	}
	
	// question 6 (1 pt)
	/**
	 * Mise en oeuvre de la partie de l'algorithme de Naimi-Tréhel qui gère le réception du jeton par
	 * un Node dans la liste d'attente.
	 * @throws IOException
	 */
	public void receivingToken() throws IOException {
		GEN.info("[" + this.id.getNum() + "] le noeud " + " reçoit le jeton et obtient la ressource ");
		havingToken = true;
		sendTaskToServer();
	}

	// question 7 1 pt
	/**
	 * Mise en oeuvre de la partie de l'algorithme de Naimi-Tréhel qui gère la libération de la ressource.
	 * @throws IOException
	 *                toutes les IOException pouvant se produire lors des échanges et des connexions.
	 */
	public void releasingResource() throws IOException {
		GEN.info("[" + this.id.getNum() + "] le noeud  libère la ressource ");
		requesting = false;
		if (nextConnexion != null) {
			GEN.info("[" + this.id.getNum() + "] le noeud  donne le jeton au noeud suivant " + next.getNum());
			nextConnexion.sendMsg(MessType.TOKENTYPE.ordinal(), new Token()); // 0.5
			havingToken = false;
			closeFullDuplexMsgWorker(nextConnexion);
			next = null;
			nextConnexion = null; // chainages close : 0.5
		}
	}
	
	
	// question 8 2 pts
	/**
	 * Ouvre le canal listenChannel pour accepter les connexions qui arriveront sur le port "myPort".
	 * Met le canal en asynchrone et l'enregistre auprès du Selector pour une gestion asynchrone des connexions.
	 * @param myPort
	 *               le port d'écoute du canal.
	 * @throws IOException
	 *               toutes les IOException qui peuvent de produire lors de la préparation de la réception des connexions.
	 */
	public void openListenChannel(final int myPort) throws IOException {
		listenChannel = ServerSocketChannel.open(); // 0.5
		listenChannel.setOption(StandardSocketOptions.SO_REUSEADDR, true);
		listenChannel.bind(new InetSocketAddress(myPort)); // 0.5
 		listenChannel.configureBlocking(false); // 0.5
		listenChannel.register(selector, SelectionKey.OP_ACCEPT); // 0.5

	}
	
	
	
	// question 9 : 5 pts
	/**
	 * mise en oeuvre de l'algorithme de Naimi-Tréhel.
	 * @throws IOException 
	 * 			Toutes les exceptions d'entrée sortie.
	 * @throws InterruptedException 
	 */
	public void algoNaimiTrehel() throws IOException, InterruptedException {

		while (true) {
			DebugTools.sendToLog(this);
			
			selector.select(); // 1 pt gestion du selector   jusqu' la boucle
			Set<SelectionKey> readyKeys = selector.selectedKeys(); 

			COMM.info("{}", () -> "[" + this.id.getNum() + "] Server: Nb Ready keys:" + readyKeys.size());


			for (SelectionKey curKey : readyKeys) { 
				if (curKey.isAcceptable()) { // 1 pt toute la gestion accept  
					// connexion d'un noeud fils de l'arbre ou
					// prédécesseur de la file d'attente ou de la console
					SocketChannel rwChan = listenChannel.accept(); 
					rwChan.configureBlocking(false);
					SelectionKey newKey = rwChan.register(selector, SelectionKey.OP_READ);
					FullDuplexMsgWorker	worker = new FullDuplexMsgWorker(rwChan);
					inConnexions.put(newKey, worker);
					GEN.info("[" + this.id.getNum() + "] le noeud  a accepté une connexion");
				} else {
					if (curKey.isReadable()) {
						// message venant soit 
						//   -> d'un fils (donc un ResourceRequest)
						//   -> d'un prédécessur (donc un Token)      
						//   -> du serveur ressource (donc un TaskResponse)
						//   -> de la console (donc un TaskResquest)
						FullDuplexMsgWorker worker = inConnexions.get(curKey);
						ReadMessageStatus status;
						status = worker.readMessage();
						if (status == ReadMessageStatus.ChannelClosed) { // closed channel

							COMM.info("[" + this.id.getNum() + "] closed channel detected");

							inConnexions.remove(curKey);
							curKey.cancel();
							closeFullDuplexMsgWorker(worker);
						} // 1 pt
						if (status == ReadMessageStatus.ReadDataCompleted) { // 0.5
							switch (MessType.values()[worker.getMessType()]) {
							case RESOURCEREQUESTTYPE: // 0.5
								ResourceRequest csRequest = (ResourceRequest) worker.getData();
								receivingResourceRequest(csRequest);
								break;
							case TOKENTYPE: // 0.5
								Token token = (Token) worker.getData();
								receivingToken();
								break;
							case TASKREQUESTTYPE: // 0.5
								this.console = worker;
								TaskRequest taskRequest = (TaskRequest) worker.getData();
								requestingResource(taskRequest);
								break;
							case TASKRESPONSETYPE: // 0.5
								TaskResponse taskResponse = (TaskResponse) worker.getData();
								releasingResource();
								console.sendMsg(MessType.TASKRESPONSETYPE.ordinal(), taskResponse);
								break;
							default:

								COMM.warn("{}", () -> "[" + this.id.getNum() + "] reception of unknown type of messages: "
										+ worker.getMessType());

							}

						}

						COMM.trace("{}", () -> "[" + this.id.getNum() + "] Received message status:" + status);


					}

				}

			}
			readyKeys.clear(); // 0.5

		}


	}
	
	//*************************** Constructeur, getters et setters pour faciliter les tests **************************/

	/**
	 * 
	 */
	public Node() {
		
	}
	/**
	 * 
	 * @return
	 *       id du noeud.
	 */
	public NodeId getId() {
		return id;
	}
	
	
	/**
	 * 
	 * @return
	 *      id du noeud pêre.
	 */
	public NodeId getFather() {
		return father;
	}

	/**
	 * 
	 * @return
	 *     id du noeud suivant.
	 */
	public NodeId getNext() {
		return next;
	}

	/**
	 * 
	 * @return
	 *      connexion vers le père.
	 */
	public FullDuplexMsgWorker getFatherConnexion() {
		return fatherConnexion;
	}

	/**
	 * 
	 * @return
	 *    connxion vers le suivant.
	 */
	public FullDuplexMsgWorker getNextConnexion() {
		return nextConnexion;
	}

	/**
	 * 
	 * @return
	 *      true si le noeud a le jeton.
	 */
	public boolean isHavingToken() {
		return havingToken;
	}

	/**
	 * 
	 * @return
	 *       true le noeud réclame de ressource critique.
	 */
	public boolean isRequesting() {
		return requesting;
	}

	/**
	 * 
	 * @return
	 *       la requête en cours pour le noeud.
	 */
	public TaskRequest getTaskRequest() {
		return taskRequest;
	}

	
	
	/**
	 * 
	 * @return
	 *       connexion vers la console.
	 */
	public FullDuplexMsgWorker getConsole() {
		return console;
	}

	/**
	 * 
	 * @return
	 *       canal serveur du noeud.
	 */
	public ServerSocketChannel getListenChannel() {
		return listenChannel;
	}

	/**
	 * 
	 * @return
	 *       le selector gérant les entrées asynchrones.
	 */
	public Selector getSelector() {
		return selector;
	}

	/**
	 * 
	 * @return
	 *       dictionnaire des connexions asynchrones avec lecture possible.
	 */
	public Map<SelectionKey, FullDuplexMsgWorker> getInConnexions() {
		return inConnexions;
	}

	/**
	 * 
	 * @return
	 *       adresse TCP/IP du serveur de tâche.
	 */
	public InetSocketAddress getServerAddr() {
		return serverAddr;
	}

	/**
	 * 
	 * @param id
	 */
	public void setId(final NodeId id) {
		this.id = id;
	}

	/**
	 * 
	 * @param father
	 */
	public void setFather(final NodeId father) {
		this.father = father;
	}

	/**
	 * 
	 * @param fatherConnexion
	 */
	public void setFatherConnexion(final FullDuplexMsgWorker fatherConnexion) {
		this.fatherConnexion = fatherConnexion;
	}

	/**
	 * 
	 * @param next
	 */
	public void setNext(final NodeId next) {
		this.next = next;
	}

	/**
	 * 
	 * @param nextConnexion
	 */
	public void setNextConnexion(final FullDuplexMsgWorker nextConnexion) {
		this.nextConnexion = nextConnexion;
	}

	/**
	 * 
	 * @param havingToken
	 */
	public void setHavingToken(final boolean havingToken) {
		this.havingToken = havingToken;
	}

	/**
	 * 
	 * @param requesting
	 */
	public void setRequesting(final boolean requesting) {
		this.requesting = requesting;
	}

	/**
	 * 
	 * @param listenChannel
	 */
	public void setListenChannel(final ServerSocketChannel listenChannel) {
		this.listenChannel = listenChannel;
	}

	/**
	 * 
	 * @param selector
	 */
	public void setSelector(final Selector selector) {
		this.selector = selector;
	}

	/**
	 * 
	 * @param inConnexions
	 */
	public void setInConnexions(final Map<SelectionKey, FullDuplexMsgWorker> inConnexions) {
		this.inConnexions = inConnexions;
	}

	/**
	 * 
	 * @param taskRequest
	 */
	public void setTaskRequest(final TaskRequest taskRequest) {
		this.taskRequest = taskRequest;
	}

	/**
	 * 
	 * @param serverAddr
	 */
	public void setServerAddr(final InetSocketAddress serverAddr) {
		this.serverAddr = serverAddr;
	}

	/**
	 * 
	 * @param console
	 */
	public void setConsole(final FullDuplexMsgWorker console) {
		this.console = console;
	}
	
}
